REM end skip pattern display at last 1 bit when
REM show.trailingzeroes = 0 and
REM show.underscores = 1

DECLARE SUB WRWARNING (a$, b$)
DECLARE SUB WRERROR (a$, b$)
DECLARE SUB PAD ()

REM Program to read P2 assembler code from input file
REM and write skip patterns to output file

REM Written by Tony Brewer, UK, 2020
REM This code may be used by anyone for any purpose free of charge

REM String constants
DIM ini.file AS STRING: REM .ini filename       
DIM input.type AS STRING: REM input file type
DIM output.type AS STRING: REM output file type
DIM SHARED comment.string AS STRING: REM comment string
DIM ignore.string AS STRING: REM ignore line string
DIM execf.string AS STRING: REM start string for EXECF
DIM skipf.string AS STRING: REM start string for SKIPF
DIM bits.string AS STRING: REM show bits string
DIM zerobits.string AS STRING: REM show zero bits string
DIM name.prefix AS STRING: REM pattern name prefix
DIM name.suffix AS STRING: REM pattern name suffix
DIM value.prefix AS STRING: REM pattern value binary prefix
DIM value.suffix AS STRING: REM pattern value binary suffix
DIM bits.prefix AS STRING: REM bit count prefix
DIM bits.suffix AS STRING: REM bit count suffix
DIM zerobits.prefix AS STRING: REM bit 0 count prefix
DIM zerobits.suffix AS STRING: REM bit 0 count suffix

REM Integer constants
DIM show.underscores AS INTEGER: REM 0 if no underscores in output file
DIM show.trailingzeroes AS INTEGER: REM 0 if no trailing 0s in output file
DIM show.statistics AS INTEGER: REM 0 if no stats in output file
DIM SHARED show.warnings AS INTEGER: REM 0 if no warnings in output file
DIM show.bits AS INTEGER: REM 0 if no bit counts in output file
DIM name.column AS INTEGER: REM pattern name starting column
DIM equal.column AS INTEGER: REM equal sign column
DIM value.column AS INTEGER: REM pattern value starting column
DIM SHARED tab.size AS INTEGER: REM tab size

REM String variables
DIM inputfile$: REM source filename
DIM outputfile$: REM skip pattern constants filename
DIM key.name AS STRING: REM .ini file key name
DIM key.value AS STRING: REM .ini file key value
REM DIM lin$ AS STRING: REM input or output line
DIM lin$: REM input or output line
DIM char AS STRING: REM single character from line
DIM labelchars AS STRING: REM label chars, order unimportant, no spaces
DIM skip.chars AS STRING: REM skip chars as used in patterns, order critical, spaces perhaps
DIM skip.name(1 TO 62) AS STRING: REM bit pattern name for labelchars(x)
DIM skip.value(1 TO 124) AS STRING: REM bit pattern value for labelchars(x)
DIM skip.bits  AS STRING: REM bit pattern with 0s or 1s only
DIM count.name(1 TO 62) AS STRING: REM bit length name for labelchars(x)
DIM count.value(1 TO 62) AS INTEGER: REM bit length value for labelchars(x)
DIM label AS STRING: REM first non-space char after execf/skipf string
DIM SHARED pads AS STRING: REM Padding for output file

REM Integer variables
DIM n, x, y, z AS INTEGER
DIM SHARED lino AS INTEGER: REM line number for input file
DIM skip.type(1 TO 62) AS INTEGER: REM 1 for SKIPF, 0 for EXECF
DIM length AS INTEGER: REM length of current pattern including _
DIM instructions AS INTEGER: REM total instructions
DIM patterns AS INTEGER: REM total patterns
DIM zeroes AS INTEGER: REM total zeroes in patterns
DIM SHARED warnings AS INTEGER: REM total warnings
DIM SHARED errors AS INTEGER: REM total warnings
DIM skip1exec0 AS INTEGER: REM 1 for SKIPF, 0 for EXECF
DIM write.zerobits AS INTEGER: REM 1 to output bit 0 counts
DIM validchar AS INTEGER: REM non-zero if skip char is valid
DIM validpattern AS INTEGER: REM non-zero if pattern block is valid
DIM SHARED warnerr AS INTEGER: REM non-zero if warning or error
DIM SHARED column AS INTEGER: REM current column
DIM SHARED new.column AS INTEGER: REM next column
DIM last1 AS INTEGER: REM string position of last 1 in skip block

REM CLS

REM Default .ini file settings
input.type = ".SP2"
output.type = ".SKP"
comment.string = "'"
ignore.string = "'#"
execf.string = "''"
skipf.string = "'''"
name.prefix = ""
name.suffix = "_"
value.prefix = "%"
value.suffix = ""
zerobits.prefix = ""
zerobits.suffix = "_0"
name.column = 1
equal.column = 17
value.column = 25
bits.column = 73
tab.size = 8
show.underscores = 1
show.trailingzeroes = 1
show.bits = 0
show.statistics = 1
show.warnings = 1

CLS

REM ON ERROR GOTO inputfile

REM Read .ini file
ini.file = "P2SKP.INI"
OPEN ini.file FOR INPUT AS #3

REM Read settings from .ini file
DO WHILE NOT EOF(3)
  LINE INPUT #3, lin$
  IF lin$ <> "" THEN
    REM Line not blank
    REM Check for equal sign
    n = INSTR(lin$, "=")
    IF n > 0 THEN
      key.name = LCASE$(LTRIM$(RTRIM$(LEFT$(lin$, n - 1))))
      key.value = LTRIM$(RTRIM$(MID$(lin$, n + 1, LEN(lin$) - n)))
      SELECT CASE key.name
	CASE "input.type"
	  input.type = key.value
	CASE "output.type"
	  output.type = key.value
	CASE "comment.string"
	  comment.string = key.value
	CASE "ignore.string"
	  ignore.string = key.value
	CASE "execf.string"
	  execf.string = key.value
	CASE "skipf.string"
	  skipf.string = key.value
	CASE "name.prefix"
	  name.prefix = key.value
	CASE "name.suffix"
	  name.suffix = key.value
	CASE "value.prefix"
	  value.prefix = key.value
	CASE "value.suffix"
	  value.suffix = key.value
	CASE "zerobits.prefix"
	  zerobits.prefix = key.value
	CASE "zerobits.suffix"
	  zerobits.suffix = key.value
	CASE "name.column"
	  name.column = VAL(key.value)
	CASE "equal.column"
	  equal.column = VAL(key.value)
	CASE "value.column"
	  value.column = VAL(key.value)
	CASE "bits.column"
	  bits.column = VAL(key.value)
	CASE "tab.size"
	  tab.size = VAL(key.value)
	CASE "show.underscores"
	  show.underscores = VAL(key.value)
	CASE "show.trailingzeroes"
	  show.trailingzeroes = VAL(key.value)
	CASE "show.bits"
	  show.bits = VAL(key.value)
	CASE "show.statistics"
	  show.statistics = VAL(key.value)
	CASE "show.warnings"
	  show.warnings = VAL(key.value)
      END SELECT
    END IF
  END IF
LOOP

inputfile:

REM ON ERROR GOTO nofile

REM read command line argument
REM ignore chars after a space
lin$ = "": REM lin$ = LTRIM$(RTRIM$(COMMAND$))
x = INSTR(lin$, " ")
IF x > 0 THEN
  lin$ = LEFT$(lin$, x - 1)
END IF

REM Prompt for input file if no command line argument
DO WHILE lin$ = ""
  INPUT "Source input file? ", lin$
LOOP
x = INSTR(lin$, ".")
IF x > 0 THEN
  inputfile$ = lin$
  outputfile$ = LEFT$(lin$, x - 1) + output.type
ELSE
  inputfile$ = lin$ + input.type
  outputfile$ = lin$ + output.type
END IF

REM Open input and output files
OPEN inputfile$ FOR INPUT AS #1
OPEN outputfile$ FOR OUTPUT AS #2
PRINT "Input  file = "; inputfile$
PRINT "Output file = "; outputfile$

lino = 0
validchar = 0
validpattern = 0
length = 0
instructions = 0
patterns = 0
zeroes = 0
warnings = 0
errors = 0
warnerr = 0
skip1exec0 = 0

labelchars = ""
skip.chars = ""
FOR n = 1 TO 124
  IF n < 63 THEN skip.name(n) = ""
  skip.value(n) = ""
NEXT

REM Loop through lines in input file
DO WHILE NOT EOF(1)

  REM Read next line
  LINE INPUT #1, lin$
  REM PRINT lin$
  lino = lino + 1

  IF lin$ = "" OR LEFT$(lin$, LEN(ignore.string)) = ignore.string THEN
    REM blank or ignore line

  REM ------------------------------------------------

  ELSE
  
    REM not blank line
    REM check for skipf string
    n = INSTR(lin$, skipf.string)
    IF n > 0 THEN

      REM skipf string found
      label = RIGHT$(lin$, LEN(lin$) - (n + LEN(skipf.string) - 1))
      skip1exec0 = 1

    ELSE

      REM no skipf string
      REM check for execf string
      n = INSTR(lin$, execf.string)
      IF n > 0 THEN

	REM execf string found
	label = RIGHT$(lin$, LEN(lin$) - (n + LEN(execf.string) - 1))
        skip1exec0 = 0

      END IF

    END IF
    
    REM n > 0 if skipf/execf found
    IF n > 1 AND LEN(lin$) > n + 1 THEN

      REM skipf/execf in line but not at start
      REM test for comment string at start
      char = LEFT$(lin$, 1)
      IF char = comment.string THEN
	REM assume label starts at 2nd char
	lin$ = RIGHT$(lin$, LEN(lin$) - 1)
	n = n - 1
	char = LEFT$(lin$, 1)
      END IF

      REM test whether char is start of label
      IF char >= "A" AND char <= "Z" OR char >= "a" AND char <= "z" THEN

	REM here if label at start of line

	REM search for space or tab after label
	x = 1
	DO WHILE x <= LEN(lin$) AND MID$(lin$, x, 1) <> " " AND MID$(lin$, x, 1) <> CHR$(9)
	  x = x + 1
	LOOP
	REM label = LEFT$(lin$, x - 1)
	REM x = no. of first char after label

	REM test char after skipf/execf
	char = LEFT$(LTRIM$(label), 1)
	IF char >= "0" AND char <= "9" OR char >= "A" AND char <= "Z" OR char >= "a" AND char <= "z" THEN

	  REM skip char valid
	  IF INSTR(labelchars, char) = 0 THEN

	    REM new skip char
	    labelchars = labelchars + char
	    n = LEN(labelchars)
	    skip.name(n) = LEFT$(lin$, x - 1)
	    skip.type(n) = skip1exec0

	  ELSE

	    REM char already used
	    WRERROR "Duplicated ", char

	  END IF
	    
	ELSE

	  REM char not alphanumeric
	  WRERROR "Not alphanumeric ", char
		
	END IF
      END IF
    
    ELSEIF n = 1 THEN

      REM skipf or execf at start of line
      REM terminate all skip patterns if started
      REM no warnings or errors for this code block
      IF validpattern <> 0 THEN

        IF MID$(lin$, LEN(skipf.string) + 1, 1) = "0" THEN
          write.zerobits = 1
        ELSE
          write.zerobits = 0
        ENDIF

	REM valid patterns were started
	patterns = patterns + LEN(labelchars)
	IF warnerr > 0 AND show.warnings <> 0 THEN
	  PRINT #2,
	END IF
	warnerr = 0

	REM *******

	last1 = 255
	REM find last 1 (actually earliest 1 due to reversal)
	FOR x = 1 TO LEN(skip.chars)
	  char = MID$(skip.chars, x, 1)
	  IF char >= "0" AND char <= "9" OR char >= "A" AND char <= "Z" OR char >= "a" AND char <= "z" THEN
	    REM alphanumeric char
	    REM check char has a label
	    n = INSTR(labelchars, char)
	    IF n > 0 THEN
	      y = INSTR(skip.value(x), "1")
	      IF y > 0 AND y < last1 THEN
		last1 = y
	      END IF
	    END IF
	  END IF
	NEXT
	IF last1 = 255 THEN
	  last1 = 1
	END IF

	REM *******

	REM write each pattern to file
	FOR x = 1 TO LEN(skip.chars)

	  count.name(x) = ""

	  char = MID$(skip.chars, x, 1)
	  IF char >= "0" AND char <= "9" OR char >= "A" AND char <= "Z" OR char >= "a" AND char <= "z" THEN

	    REM alphanumeric char
	    REM check char has a label
	    n = INSTR(labelchars, char)
	    IF n > 0 THEN

	      REM n = labelchars/skip.name index
	      REM x = skip.chars/skip.value index

	      IF write.zerobits <> 0 THEN
		count.name(x) = zerobits.prefix + skip.name(n) + zerobits.suffix
		REM Count 0s in skip.value(x)
		z = 0
		FOR y = 1 TO LEN(skip.value(x))
		  IF MID$(skip.value(x), y, 1) = "0" THEN
		    z = z + 1
		  END IF
		NEXT
		count.value(x) = z
	      END IF

	      REM change space to _ in skip.value(x)
	      FOR y = 1 TO LEN(skip.value(x))
		IF MID$(skip.value(x), y, 1) = " " THEN
		  MID$(skip.value(x), y, 1) = "_"
		END IF
	      NEXT

	      REM *** change below ***

	      REM test for underscore output
	      IF show.underscores = 0 AND show.trailingzeroes = 0 THEN
		REM no underscores
		temp$ = ""
		FOR y = 1 TO LEN(skip.value(x))
		  char = MID$(skip.value(x), y, 1)
		  IF char <> "_" THEN
		    temp$ = temp$ + char
		  END IF
		NEXT
		REM test for null string
		IF temp$ = "" THEN
		  skip.value(x) = "0"
		ELSE
		  skip.value(x) = temp$
		END IF
	      END IF

	      REM test for trailing zeroes output
	      IF show.trailingzeroes = 0 THEN
		IF show.underscores = 0 THEN
		  REM no trailings 0s at all
		  WHILE LEFT$(skip.value(x), 1) = "0"
		    y = LEN(skip.value(x))
		    skip.value(x) = RIGHT$(skip.value(x), y - 1)
		  WEND
		  IF y = 1 THEN
		    skip.value(x) = "0"
		  END IF
		ELSE
		  REM no trailings 0s after last 1
		  skip.value(x) = RIGHT$(skip.value(x), LEN(skip.value(x)) - (last1 - 1))
		END IF
	      END IF

	      REM *** change above ***

	      REM compose new line as Name = Value at specified columns
	      column = 0
	      new.column = name.column
	      PAD
	      lin$ = name.prefix + skip.name(n) + name.suffix
	      column = name.column + LEN(lin$) - 1
	      lin$ = pads + lin$
	      new.column = equal.column
	      PAD
	      lin$ = lin$ + pads + "="
	      column = column + 1
	      new.column = value.column
	      PAD
	      lin$ = lin$ + pads + value.prefix + skip.value(x) + value.suffix
	      column = column + LEN(value.prefix + skip.value(x) + value.suffix) - 1
	      IF tab.size = 0 THEN
		column = column + 1: REM kludge
	      END IF

	      IF show.bits <> 0 THEN
		new.column = bits.column
		PAD
		REM IF skip1exec0 = 1 THEN
		REM   lin$ = lin$ + pads + skipf.string
		REM ELSE
		REM   lin$ = lin$ + pads + execf.string
		REM END IF
		lin$ = lin$ + pads + comment.string
		REM find number of 0s or 1s in skip.value(x)
		z = 0
		FOR y = 1 TO LEN(skip.value(x))
		  IF MID$(skip.value(x), y, 1) = "0" OR MID$(skip.value(x), y, 1) = "1" THEN
		    z = z + 1
		  END IF
		NEXT
		IF z < 10 THEN
		  lin$ = lin$ + " "
		END IF
		lin$ = lin$ + STR$(z)
	      END IF

	      REM write new line to output file
	      PRINT #2, lin$
	      skip.name(n) = ""
	      skip.value(x) = ""
	    ELSE
	      REM no label for char
	      REM ignore, error already written
	    END IF
	  ELSE
	    REM not alphanumeric char
	    REM ignore, no error
	  END IF
	NEXT

	REM write bit lengths to file
	IF write.zerobits <> 0 THEN
	  PRINT #2,
	  FOR x = 1 TO LEN(skip.chars)
	    IF count.name(x) <> "" THEN

	      'lin$ = count.name(x) + " = " + LTRIM$(STR$(count.value(x)))
	      'PRINT #2,lin$
'****
	      REM compose new line as Name = Value at specified columns
	      column = 0
	      new.column = name.column
	      PAD
	      lin$ = count.name(x)
	      column = name.column + LEN(lin$) - 1
	      lin$ = pads + lin$
	      new.column = equal.column
	      PAD
	      lin$ = lin$ + pads + "="
	      column = column + 1
	      new.column = value.column
	      PAD
	      lin$ = lin$ + pads + LTRIM$(STR$(count.value(x)))
	      PRINT #2, lin$

	      'column = column + LEN(value.prefix + skip.value(x) + value.suffix) - 1
	      'IF tab.size = 0 THEN
	      '  column = column + 1 : REM kludge
	      'END IF
'****

	    END IF
	  NEXT
	END IF

	PRINT #2,
	validpattern = 0
	skip1exec0 = 0
	length = 0
	write.zerobits = 0
	labelchars = ""
	skip.chars = ""
      END IF
    
    REM ------------------------------------------------

    ELSE

      REM Here if no skipf/execf start string in line
      REM Test for skip pattern if labelchars not null string
      validchar = 0
      IF labelchars <> "" THEN

	REM Remove leading spaces and tabs
	lin$ = LTRIM$(lin$)
	DO WHILE LEFT$(lin$, 1) = CHR$(9)
	  lin$ = RIGHT$(lin$, LEN(lin$) - 1)
	LOOP

	REM If comment string at start of line remove it and look for another
	x = LEN(lin$)
	n = INSTR(lin$, comment.string)
	IF n = 1 AND x > 1 THEN
	  lin$ = RIGHT$(lin$, x - 1)
	  n = INSTR(lin$, comment.string)
	END IF

	IF n > 1 AND LEN(lin$) > n THEN

	  REM comment string in line, assumed to be after valid code
	  REM check for ## twice before comment string
	  repeat = 1: REM assume no ##
	  FOR y = 2 TO 3
	    x = INSTR(lin$, "##")
	    IF x > 1 AND n > x THEN
	      repeat = y
	      lin$ = MID$(lin$, x + 1, LEN(lin$) - x)
	    END IF
	  NEXT

	  REM crop line after comment string
	  lin$ = MID$(lin$, n + 1, LEN(lin$) - n)

	  REM assume cropped line starts with skip pattern string and
	  REM crop up to first char that is not space, 0-9, A-Z, a-z, or |
	  n = LEN(lin$)
	  IF n > 0 THEN
	    x = 0
	    DO
	      x = x + 1
	      y = x
	      char = MID$(lin$, x, 1)
	      IF NOT (char = " " OR char >= "0" AND char <= "9" OR char >= "A" AND char <= "Z" OR char >= "a" AND char <= "z" OR char = "|") THEN
		x = x - 1
	      END IF
	    LOOP UNTIL x = n OR x <> y
	  END IF
	  lin$ = LEFT$(lin$, x)

	  REM lin$ is null string or skip pattern
	  n = LEN(lin$)
	  IF n > 0 THEN

	    REM lin$ is new skip pattern
	    FOR x = 1 TO n

	      REM get next char
	      char = MID$(lin$, x, 1)

	      IF x > LEN(skip.chars) THEN
		REM extend skip string
		skip.chars = skip.chars + char
	      END IF

	      REM set label to unknown
	      label$ = "?"

	      REM test for space, | or alphanumeric char

	      IF char = " " AND skip.value(x) <> "" THEN
		REM add space
		FOR y = 1 TO repeat
		  skip.value(x) = " " + skip.value(x)
		NEXT

	      ELSEIF char = "|" THEN

		REM set label if skip char defined
		IF LEN(skip.chars) >= x THEN
		  y = INSTR(labelchars, MID$(skip.chars, x, 1))
		  IF y > 0 THEN
		    label$ = skip.name(y)
		    skip1exec0 = skip.type(y)
		  END IF
		END IF

		REM pipe char, equivalent to 1
		validchar = 1

		REM test for 1 and adding underscores at start
		IF skip.value(x) = "" AND length > 0 THEN
		  FOR y = 1 TO length
		    skip.value(x) = "_" + skip.value(x)
		  NEXT
		END IF

		REM test for 1 after space
		IF LEFT$(skip.value(x), 1) = " " THEN
		  WRWARNING "1 after space for ", label$
		END IF

		REM add 1(s)
		validpattern = 1
		FOR y = 1 TO repeat
		  skip.value(x) = "1" + skip.value(x)
		NEXT

		REM set skip.bits to 0 or 1 only
		skip.bits = ""
		FOR y = 1 TO LEN(skip.value(x))
		  char = MID$(skip.value(x), y, 1)
		  IF char = "0" OR char = "1" THEN
		    skip.bits = skip.bits + char
		  END IF
		NEXT

		REM test for 1 after limit or big jump
		y = INSTR(skip.bits, "1")
		z = LEN(skip.bits)
		IF y > 0 THEN

		  REM 1 found in pattern
		  REM test for 1 after 22 or 32 bits

		  IF skip1exec0 = 0 AND z - y > 22 OR skip1exec0 = 1 AND z - y > 32 THEN
		    WRERROR "1 beyond limit for ", label$
                    STOP
		  END IF
 
		  REM test for "11111111"
		  y = INSTR(skip.bits, "11111111")
		  char = MID$(skip.value(x), 9, 1)
		  REM warn if jump = 8, not > 8
		  IF y = 1 AND char <> "1" THEN
		    WRWARNING "Jump > 7 for ", label$
		  END IF
		  
		END IF

	      ELSEIF char >= "0" AND char <= "9" OR char >= "A" AND char <= "Z" OR char >= "a" AND char <= "z" THEN

		REM alphanumeric char

		REM REM check char not used earlier in pattern string
		REM IF INSTR(lin$, char) = x THEN

		  REM test for label
		  y = INSTR(labelchars, char)
		  IF y > 0 THEN

		    REM label defined
		    label$ = skip.name(y)

		    REM check for underscores at start
		    IF skip.value(x) = "" AND length > 0 THEN
		      REM insert underscores
		      FOR y = 1 TO length
			skip.value(x) = "_" + skip.value(x)
		      NEXT
		    END IF

		    REM check for matching char or pipe
		    IF MID$(skip.chars, x, 1) = char OR MID$(skip.chars, x, 1) = " " OR MID$(skip.chars, x, 1) = "|" THEN

		      MID$(skip.chars, x, 1) = char
		      validchar = 1
		      zeroes = zeroes + 1

		      REM test for 1 at start (EXECF only)
		      IF INSTR(skip.value(x), "0") = 0 AND INSTR(skip.value(x), "1") > 0 AND skip1exec0 = 0 THEN
			WRWARNING "1 at start for ", label$
		      END IF

		      REM test for 0 after space
		      IF LEFT$(skip.value(x), 1) = " " THEN
			WRWARNING "0 after space for ", label$
		      END IF

		      REM add zero(es)
		      validpattern = 1
		      FOR y = 1 TO repeat
			skip.value(x) = "0" + skip.value(x)
		      NEXT

		    ELSE
		      REM char mismatch
		      WRERROR "Mismatch for ", char
		    END IF
		  
		  ELSE
		    REM no label
		    WRERROR "No label for ", char
		  END IF

		REM ELSE
		REM   REM char already used
		REM   WRERROR "Duplicated ", char
		REM END IF

	      ELSE
		REM not space, pipe or alphanumeric char
		REM ignore
	      END IF

	    NEXT

	    IF validchar = 1 THEN
	      length = length + repeat
	      instructions = instructions + repeat
	    END IF

	  END IF
	END IF
      END IF
    END IF
  END IF
LOOP

REM Write statistics if enabled to end of output file
IF show.statistics <> 0 THEN
  REM stats enabled
  lin$ = comment.string + " " + LTRIM$(STR$(patterns)) + " skip patterns with "
  lin$ = lin$ + LTRIM$(STR$(zeroes)) + " zero bits using "
  lin$ = lin$ + LTRIM$(STR$(instructions)) + " instruction longs"
  PRINT #2, lin$
  lin$ = comment.string + " Code compression = " + LTRIM$(STR$(zeroes)) + "/"
  IF instructions > 0 THEN
    lin$ = lin$ + LTRIM$(STR$(instructions)) + " = " + LTRIM$(STR$(zeroes / instructions))
  END IF
  REM show only 3 decimal places
  x = INSTR(lin$, ".")
  IF x > 0 AND LEN(lin$) > x + 3 THEN
    lin$ = LEFT$(lin$, x + 3)
  END IF
  PRINT #2, lin$
  PRINT #2, comment.string; " Warnings = "; LTRIM$(STR$(warnings))
  PRINT #2, comment.string; " Errors = "; LTRIM$(STR$(errors))
END IF

PRINT "Warnings ="; warnings
PRINT "Errors ="; errors

REM End program, closing all files
END

nofile:
PRINT "Input file not found"
END

SUB PAD
  DIM m, n, x, y AS INTEGER
  pads = ""
  m = new.column - column
  IF m <= 1 THEN
    REM Here if new.column < column + 2
    IF new.column > 1 OR (m < 1 AND column > 1) THEN
      pads = " "
      column = column + 1
    END IF
  ELSEIF tab.size = 0 THEN
    REM Pad with spaces only
    FOR n = 1 TO m - 1
      pads = pads + " "
      column = column + 1
    NEXT
  ELSE
    REM Pad with tabs and spaces
    x = column - 1
    y = new.column - 1
    n = y - y MOD tab.size - (x - x MOD tab.size)
    IF x MOD tab.size = 7 THEN
      n = n - tab.size
    END IF
    IF n >= tab.size THEN
      WHILE n >= tab.size
	pads = pads + CHR$(9)
	n = n - tab.size
      WEND
      WHILE n > 0
	pads = pads + " "
	n = n - 1
      WEND
    ELSE
      n = new.column - column
      WHILE n > 0
	pads = pads + " "
	n = n - 1
      WEND
    END IF
    column = new.column
  END IF
END SUB

SUB WRERROR (a$, b$)
  warnerr = 1
  errors = errors + 1
  PRINT #2, comment.string; " *Error* in line "; LTRIM$(STR$(lino)); " : "; a$; b$
END SUB

SUB WRWARNING (a$, b$)
  warnerr = 1
  warnings = warnings + 1
  IF show.warnings <> 0 THEN
    PRINT #2, comment.string; " Warning in line "; LTRIM$(STR$(lino)); " : "; a$; b$
  END IF
END SUB

